import asyncio
from typing import Callable

class Tweens:

    tweens: list = []

    def __init__(self):
        self.tweens = []

    def add(self, target: any,
            property: str,
            fr: float,
            to: float,
            repeat: int = 0,
            duration: int = 1000,
            delay: int = 0,
            after_tween: Callable | None = None,
            update_tween: Callable | None = None):
        loop = asyncio.get_event_loop()
        loop.create_task(self.start_tween(target, property, fr, to, repeat, duration, delay, after_tween, update_tween))
        
    async def start_tween(self, target: any,
            property: str | None = None,
            fr: float | None = None,
            to: float | None = None,
            repeat: int = 0,
            duration: int = 1000,
            delay: int = 0,
            after_tween: Callable | None = None,
            update_tween: Callable | None = None):
        await asyncio.sleep(delay)
        cnt = (1, repeat + 1)[repeat != -1]
        while cnt > 0:
            for i in range(60 * duration):
                if update_tween:
                    update_tween()
                else:
                    setattr(target, property, fr + (to - fr) * (i / (60 * duration)))
                await asyncio.sleep(1 / 60)

            if repeat != -1:
                cnt -= 1

        if after_tween:
            after_tween()
 



from typing import Callable
import pygame

from ..vmath import Vector2d
from ..game_object import Component
from ..globals import camera, DEBUG

class Clicked:
    def __init__(self, color: tuple[int, int, int], border: "Border | None" = None):
        self.color = color
        self.border = border

    @staticmethod
    def get_default():
        return Clicked((0, 0, 0))


class Border:

    def __init__(self, width: int, color: tuple[int, int, int], radius: int):
        self.width = width
        self.color = color
        self.radius = radius

    @staticmethod
    def get_default():
        return Border(1, (0, 0, 0), -1)


class Text:

    def __init__(self, text: str,
                 font: pygame.font.Font | None = None,
                 color: tuple[int, int, int] = (0, 0, 0),
                 background_color: tuple[int, int, int] | None = None):
        font = font or pygame.font.SysFont("Arial", 20)
        self.text = text
        self.font = font
        self.color = color
        self.background_color = background_color

    def render(self):
        return self.font.render(self.text, False, self.color, self.background_color)
    

class LabelComponent(Component):

    def __init__(self, text: str, color: tuple[int, int, int]):
        self.font = pygame.font.SysFont("Arial", 50) # TODO: choosing fonts
        self.text = text
        self.scale_x = 1
        self.scale_y = 1
        self.color = color
        
    def draw(self, display: pygame.Surface):
        sum_height = 0
        for i in self.text.split("\n"):
            # TODO: Rework it so center of whole passage will be at right center.
            text = self.font.render(i, True, self.color)

            text = pygame.transform.scale(text, (text.get_width() * self.scale_x, text.get_height() * self.scale_y))
            display.blit(text, text.get_rect(center=
                (self.game_object.transform.position + Vector2d(0, sum_height)).as_tuple()))
            sum_height += text.get_height()


class ButtonComponent(Component):

    def __init__(self, cmd: Callable, interception: Callable[[Vector2d, Vector2d], bool], *args):
        self.cmd = cmd
        self.interception = interception
        self.args = args

    def update(self):
        if pygame.mouse.get_pressed(3)[0]:
            pos = Vector2d.from_tuple(pygame.mouse.get_pos())
            if self.interception(camera.normalize(self.game_object.transform.position), pos):
                self.cmd(self.args)


class RectButtonComponent(ButtonComponent):

    def __init__(self, cmd: Callable, size: Vector2d, *args):
        self.size = size
        def interception(center: Vector2d, position: Vector2d) -> bool:
            temp = (center - position).operation(size, lambda a, b: -b/2 <= a <= b/2)
            return bool(temp.x) and bool(temp.y)

        super().__init__(cmd, interception)

    def draw(self, display: pygame.Surface):
        if DEBUG:
            top_left = (self.game_object.transform.position - self.size / 2)
            pygame.draw.rect(display, (200, 200, 200), (top_left.x, top_left.y, self.size.x, self.size.y), width=1)

class CircleButtonComponent(ButtonComponent):

    def __init__(self, cmd: Callable, radius: float, *args):
        self.radius = radius
        def interception(center: Vector2d, position: Vector2d) -> bool:
            return (center - position).norm() <= radius

        super().__init__(cmd, interception)


from .game_object import GameObject


class Scene:

    def __init__(self):
        pass

    def load(self, data: any):
        pass

    def destroy(self):
        for i in GameObject.objects:
            GameObject.destroy(i)

        GameObject.objects = []



import asyncio
from .scene import Scene
from .game_object import GameObject

import globals
import pygame as pg
import logging


class Engine:

    instance = None

    logger: logging.Logger
    scene: Scene
    display: pg.Surface
    fps: int

    def __init__(self, app_name: str, fps: int, resolution: tuple[int, int] = (800, 600)):
        self.fps = fps
        self.app_name = app_name
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("[%(levelname)s][%(name)s]: %(message)s"))
        self.logger.addHandler(handler)
        self.init(resolution)

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Engine, cls).__new__(cls)
        return cls.instance

    def init(self, resolution: tuple[int, int] = (800, 600)):
        pg.init()
        pg.font.init()
        logging.basicConfig()
        self.display = pg.display.set_mode(resolution)
        pg.display.set_caption(self.app_name)
        logging.info("Game initialized.")

    async def run(self, scene: Scene, data: any):
        if scene:
            scene.destroy()
            
        self.scene = scene
        scene.load(data)
        
        try:
            while True:
                await self.iteration()
                await asyncio.sleep(1 / self.fps)
        except Exception as e:
            logging.critical(e.with_traceback(None))
            exit(1)

    def event_processing(self, event: pg.event.Event):
        if event.type == pg.QUIT:
            logging.info("Quitting")
            exit()

    async def iteration(self):
        
        globals.events = pg.event.get()
        
        for event in globals.events:
            self.event_processing(event)
        
        for i in GameObject.objects:
            i.update()

        self.display.fill((0, 0, 0))
        for i in GameObject.objects:
            i.draw(self.display)

        pg.display.flip()
            



import pygame
import os

from ..globals import camera
from ..game_object import Component
from pygame import Surface

DEBUG = True

class SpriteComponent(Component):
    path: str = ''
    loaded: dict = {}

    def __init__(self, sprite_name: str, size: tuple[int, int]) -> None:
        super().__init__()

        if sprite_name in SpriteComponent.loaded.keys():
            self.texture = SpriteComponent.loaded[sprite_name]
        else:
            self.texture = pygame.image.load(os.path.join(SpriteComponent.path, sprite_name)).convert_alpha()
            SpriteComponent.loaded[sprite_name] = self.texture

        self.size = size
        self.opacity = 255

    @staticmethod
    def set_path(path: str):
        SpriteComponent.path = path

    def draw(self, display: Surface):
        self.texture.set_alpha(self.opacity)
        blit_image = self.texture

        cropped = pygame.Surface(self.size)
        cropped.blit(blit_image, (0, 0))

        angle = self.game_object.transform.angle.get()
        scale = self.game_object.transform.scale

        if angle != 0:
            cropped = pygame.transform.rotate(cropped, angle)

        if scale != 1:
            cropped = pygame.transform.scale_by(cropped, scale)

        coords = camera.normalize(self.game_object.transform.position)
        rect = cropped.get_rect(center=coords)

        if DEBUG:
            pygame.draw.rect(display, (255, 0, 0), rect, 1)
        display.blit(blit_image, rect)
    


from .vector import *


def is_in_box(vec1: Vector2d, other1: "Vector2d", other2: "Vector2d") -> bool:
    return ((vec1 - other1) * (other2 - other1)).get_quarter() == 1 and (
            (other2 - other1) * (other2 - other1) - (vec1 - other1) * (vec1 - other1)).get_quarter() == 1

def complex_multiply(vec1: Vector2d, other: Vector2d) -> Vector2d:
    return Vector2d(vec1.x * other.x - vec1.y * other.y, vec1.y * other.x + vec1.x * other.y)